#pragma once
#include <fstream>
#include <iostream>
#include <string>
#include <cstdlib>
#include <iterator>
#include <algorithm>
#include <random>
#include <functional>

//1.  createData(),        100        50  50.
void createData_v1(const std::string& fname); //,      .


template <typename Cont_T> //,   generate    
						   //   copy          .
void createData_v2(const std::string& fname) {
	std::ofstream f(fname);
	if (!f) { // file opened ?
		std::cerr << "Error: cant open output file " << fname << std::endl;
		exit(EXIT_FAILURE);
	}
	std::random_device rnd_device;
	std::mt19937 mersenne_engine(rnd_device());
	std::uniform_int_distribution<int> dist(-50, 50);
	auto gen = std::bind(dist, mersenne_engine);

	using Val_T = typename Cont_T::value_type;
	Cont_T c(100);
	std::generate(c.begin(), c.end(), gen);

	std::ostream_iterator<Val_T> output(f, " ");
	copy(c.begin(), c.end(), output);
	f.close();
};

//2.   loadData(),       .
template <typename Cont_T>
Cont_T loadData_v1(const std::string& fname) { //   2.1.,      .
	using Val_T = typename Cont_T::value_type;
	Cont_T res;
	std::ifstream f(fname);
	if (!f) { // file opened ?
		std::cerr << "Error: cant open input file " << fname << std::endl;
		exit(EXIT_FAILURE);
	}
	Val_T value;
	while (f >> value) {
		res.push_back(value);
	}
	f.close();
	return res;
}

template <typename Cont_T>
Cont_T loadData_v2(const std::string& fname) { // 2.2.,         copy.
	using Val_T = typename Cont_T::value_type;
	Cont_T res;
	std::ifstream f(fname);
	if (!f) { // file opened ?
		std::cerr << "Error: cant open input file \"" << fname << "\"" << std::endl;
		exit(EXIT_FAILURE);
	}
	std::copy(std::istream_iterator<Val_T>(f), std::istream_iterator<Val_T>(), back_inserter(res));
	f.close();
	return res;
}

//3.  modify(),        : 
//            

template <typename Cont_T>
void modify_v1(Cont_T& vec) { //	 3.1.  ,     .
							  //              .
	using Val_T = typename Cont_T::value_type;
	double ar_mean = 0;
	Val_T sum = 0;
	for (const auto& num : vec) {
		sum += num;
	}
	ar_mean = sum / vec.size();
	for (auto& num : vec) {
		if ((int)(abs(num)) % 2 == 1) num /= ar_mean;
	}
	vec.push_back(sum);
	vec.push_back(ar_mean);
}

template <typename Cont_T>
void modify_v2(const Cont_T& it1, const Cont_T& it2) {
	using Val_T = typename Cont_T::value_type;
	double ar_mean = 0;
	Val_T sum = 0;
	int n = 0;
	for (auto num = it1; num != it2; ++num) {
		sum += *num;
		n++;
	}
	ar_mean = sum / n;
	for (auto num = it1; num != it2; ++num) {
		if ((int)abs(*num) % 2 == 1) *num /= ar_mean;
	}
}

template <typename Cont_T> //3.3. ,   for_each().
						   //         - .
void modify_v3(Cont_T& vec) {
	using Val_T = typename Cont_T::value_type;
	double ar_mean = 0;
	Val_T sum = 0;
	for (const auto& num : vec) {
		sum += num;
	}
	ar_mean = sum / vec.size();
	std::for_each(vec.begin(), vec.end(), [&](auto& n) {
		if ((int)abs(n) % 2 == 1) n /= ar_mean;
		return n; });
}

void modify_v4(const std::string& fname);  //3.4. ,   transform()   
									      //     ,       .

//4.  outputResult()    - .
template <typename Cont_T> //   4.1.,      .
void outputResult_v1(Cont_T& v) {
	using Val_T = typename Cont_T::value_type;
	std::string fname = "new_data.txt";
	std::ofstream f(fname);
	if (!f) { // file opened ?
		std::cerr << "Error: cant open input file \"" << fname << "\"" << std::endl;
		exit(EXIT_FAILURE);
	}
	for (const auto& num : v) {
		f << num << " ";
	}
	f.close();
}

template <typename Cont_T> //4.2.,   copy         .
void outputResult_v2(Cont_T& v) {
	using Val_T = typename Cont_T::value_type;
	std::string fname = "new_data.txt";
	std::ofstream f(fname);
	if (!f) { // file opened ?
		std::cerr << "Error: cant open input file \"" << fname << "\"" << std::endl;
		exit(EXIT_FAILURE);
	}
	std::copy(v.begin(), v.end(), std::ostream_iterator<Val_T>(f, " "));
	f.close();
}

template <typename Cont_T>
void Application(Cont_T& cont, const std::string& fname) {
	int key;
	std::cout << "\nStep 2: Loading\n" << "Choose the method:\n";
	std::cout << "1. Using iostreams\n";
	std::cout << "2. Using stream iterators and algorithm copy\n";
	std::cout << "Another answer - using iostreams, next step\n";

	std::cin >> key;
	switch (key) {
	case(1):
		cont = loadData_v1<Cont_T>(fname);
		std::cout << "Loaded data:\n";
		for (const auto& num : cont) {
			std::cout << num << " ";
		}
		break;
	case(2):
		cont = loadData_v2<Cont_T>(fname);
		std::cout << "Loaded data:\n";
		for (const auto& num : cont) {
			std::cout << num << " ";
		}
		break;
	default:
		cont = loadData_v1<Cont_T>(fname);
		break;
	}

	std::cout << "\n\nStep 3: Modifying\n" << "Choose the method:\n";
	std::cout << "1. Function that accepts container\n";
	std::cout << "2. Function that accepts interval\n";
	std::cout << "3. Using for_each\n";
	std::cout << "4. Using transform\n";
	std::cout << "Another answer - next step\n";

	std::cin >> key;
	switch (key) {
	case(1):
		modify_v1(cont);
		std::cout << "Modifyed data:\n";
		for (const auto& num : cont) {
			std::cout << num << " ";
		}
		break;
	case(2):
		modify_v2(cont.begin(), cont.end());
		std::cout << "Modifyed data:\n";
		for (const auto& num : cont) {
			std::cout << num << " ";
		}
		break;
	case(3):
		modify_v3(cont);
		std::cout << "Modifyed data:\n";
		for (const auto& num : cont) {
			std::cout << num << " ";
		}
		break;
	case(4):
		modify_v4(fname); break;
	default:
		break;
	}

	std::cout << "\n\nStep 4: Output\n" << "Choose the method:\n";
	std::cout << "1. Using iostreams\n";
	std::cout << "2. Using copy\n";
	std::cout << "Another answer - finish\n";

	std::cin >> key;
	switch (key) {
	case(1):
		outputResult_v1(cont); break;
	case(2):
		outputResult_v2(cont); break;
	default:
		break;
	}
}